data Op  = Add | Sub | Mul | Div
 deriving (Eq,Show)
data Expr = Val Int | App Op Expr Expr
 deriving (Eq,Show)

oplist :: [Op]
oplist = [Add,Sub,Mul,Div]

subs :: [a] -> [[a]]

subs [] = [[]]
subs (x:xs) =  (subs xs) ++ map (x:) (subs xs)

inter x [] = [[x]]
inter x (y:ys) = (x:y:ys):(map (y:) (inter x ys))

perm [] = [[]]
perm (x:xs) =  concatMap (inter x) (perm xs)

choices :: [a] -> [[a]]
choices l = concatMap perm (subs l)

exprs :: [Int] -> [Expr]
exprs [] = []
exprs [x] = [Val x]
exprs l   = [ App op e1 e2 | op <- oplist, i <- [1..(length l)-1], 
                             e1 <- exprs (take i l), e2 <- exprs (drop i l) ]

evaluate :: Expr -> [(Int,Expr)]
evaluate (Val x) = [(x,Val x)]
evaluate (App Add x y) = [(fst v1 + fst v2,App Add x y) | v1 <- evaluate x, v2 <- evaluate y]
evaluate (App Mul x y) = [(fst v1 * fst v2,App Mul x y) | v1 <- evaluate x, v2 <- evaluate y]
evaluate (App Sub x y) = [(fst v1 - fst v2,App Sub x y) | v1 <- evaluate x, v2 <- evaluate y, (fst v1) >= (fst v2)]
evaluate (App Div x y) = [(div (fst v1) (fst v2),App Div x y)  | v1 <- evaluate x, v2 <- evaluate y, ((fst v2) /= 0),(mod (fst v1) (fst v2) == 0)]

solve ::  [Int] -> Int -> String
solve l v = printexp (snd (head (filter (f v)  (concatMap evaluate (concatMap exprs (choices l))))))

f v (x,y)  =  (x == v)

printexp :: Expr -> String
printexp (Val x) = show x
printexp (App Add x y) = "(" ++ (printexp x) ++ "+" ++ (printexp y) ++")" 
printexp (App Sub x y) = "(" ++ (printexp x) ++ "-" ++ (printexp y) ++")" 
printexp (App Mul x y) = "(" ++ (printexp x) ++ "*" ++ (printexp y) ++")" 
printexp (App Div x y) = "(" ++ (printexp x) ++ "/" ++ (printexp y) ++")" 
